; drv16F887\math.inc
; quelques macros pour effectuer des oprations sur 24 bits
; version 20080429

; NOTE : LE LSByte EST A LA PREMIERE ADRESSE, LE MSB EST A LA TROISIEME.

#define LOWBYTE(I)   (  I      & 255)
#define MIDBYTE(I)   (( I >> 8)& 255)
#define HIGHBYTE(I)  (( I >>16)& 255)


; utile pour initialiser des "constantes" en RAM
SET_CONSTANTE24B macro adr, val
	if (LOWBYTE(val) == 0)
		clrf	adr
    else
		movlw	LOWBYTE(val)
  		movwf	adr
    endif
	if (MIDBYTE(val) == 0)
		clrf	adr+1
    else
		movlw	MIDBYTE(val)
  		movwf	adr+1
    endif
	if (HIGHBYTE(val) == 0)
		clrf	adr+2
    else
		movlw	HIGHBYTE(val)
  		movwf	adr+2
    endif
	endm


COPY24	macro	source, dest
	movfw	source
	movwf	dest
	movfw	source+1
	movwf	dest+1
	movfw	source+2
	movwf	dest+2
	endm

;;;;;;;;;;;;;;;;; ADDITION ;;;;;;;;;;;;;;;;;

; Version avec valeur immdiate

add24imm	macro IMM, reg
	local label1
	if (LOWBYTE(IMM) != 0)
		movlw	LOWBYTE(IMM)
		addwf	reg,f
		btfss	STATUS, C
		goto	label1
			incf	reg+1,f
			btfsc	STATUS, Z
				incf	reg+2,f
label1
    endif
	if (MIDBYTE(IMM) != 0)
		movlw	MIDBYTE(IMM)
		addwf	reg+1,f
		btfsc	STATUS, C
			incf	reg+2,f
    endif
	if (HIGHBYTE(IMM) != 0)
		movlw	HIGHBYTE(IMM)
		addwf	reg+2,f
    endif
	endm

; Version avec 1 registre et un Indirect
; reg = reg - [FSR]
; FSR pointe vers le LSB de l'addende
; reg est l'adresse du LSB de la source/destination
;      /!\ Ne pas oublier de sauver FSR avant de
;         lui donner l'adresse de l'addende !!!
add24Ir		macro reg
	local label_carry
	; LSB
	movfw	INDF
	addwf	reg,f
	btfss	STATUS, C
	goto	label_carry
		incf	reg+1, f
		btfsc	STATUS, Z
			incf	reg+2,f
label_carry
	incf	FSR,f
	; MID
	movfw	INDF
	addwf	reg+1,f
	btfsc	STATUS, C
		incf	reg+2,f
	incf	FSR,f
	; MSB
	movfw	INDF
	addwf	reg+2,f
	endm


;;;;;;;;;;;;;;;;; SOUSTRACTION ;;;;;;;;;;;;;;;;;

; /!\ Comme le calcul commence par les MSB,
;  FSR pointe vers le MSB de la valeur  enlever


; version avec 1 registre et un Indirect
; reg = reg - [FSR]
; reg est l'adresse du LSB de la source/destination
sub24Ir	macro reg
	local	no_carry_sub24

	; MSB
	movfw	INDF
	subwf	reg+2,f

	; MID
	decf	FSR,f
	movfw	INDF
	subwf	reg+1,f
	btfss	STATUS, C
		decf	reg+2,f

	; LSB
	decf	FSR,f
	movfw	INDF
	subwf	reg,f
	btfsc	STATUS,C
	goto	no_carry_sub24
		movf	reg+1,f	
		btfsc	STATUS,Z
			decf	reg+2,f
		decf	reg+1,f
no_carry_sub24
	endm


; version avec 2 registres
; reg = reg - src
; src est l'adresse du LSB de la valeur  enlever  reg
; reg est l'adresse du LSB de la source/destination
sub24rr	macro reg, src
	local	no_carry_sub24

	; MSB
	movfw	src+2
	subwf	reg+2,f

	; MID
	movfw	src+1
	subwf	reg+1,f
	btfss	STATUS, C
		decf	reg+2,f

	; LSB
	movfw	src
	subwf	reg,f
	btfsc	STATUS,C
	goto	no_carry_sub24
		movf	reg+1,f
		btfsc	STATUS,Z
			decf	reg+2,f
		decf	reg+1,f
no_carry_sub24
	endm


; version avec 2 registres et un Indirect
; (utilis aussi pour les comparaisons : voir le MSBit de dest)
; dest = reg - [FSR]
; reg est l'adresse du LSB de la source
; dest est l'adresse du LSB de la destination
sub24Irr	macro reg, dest
	local	no_carry_sub24

	; MSB
	movfw	INDF
	subwf	reg+2,w
	movwf	dest+2

	; MID
	decf	FSR,f
	movfw	INDF
	subwf	reg+1,w
	movwf	dest+1
	btfss	STATUS, C
		decf	dest+2,f

	; LSB
	decf	FSR,f
	movfw	INDF
	subwf	reg,w
	movwf	dest
	btfsc	STATUS,C
	goto	no_carry_sub24
		movf	dest+1,f
		btfsc	STATUS,Z
			decf	dest+2,f
		decf	dest+1,f
no_carry_sub24

	endm


; version avec 3 registres
; (utilis aussi pour les comparaisons : voir le MSBit de dest)
; dest = reg - src
; reg est l'adresse du LSB de la source
; dest est l'adresse du LSB de la destination
sub24rrr	macro reg, dest, src
	local	no_carry_sub24
	; MSB
	movfw	src+2
	subwf	reg+2,w
	movwf	dest+2

	; MID
	movfw	src+1
	subwf	reg+1,w
	movwf	dest+1
	btfss	STATUS, C
		decf	dest+2,f

	; LSB
	movfw	src
	subwf	reg,w
	movwf	dest
	btfsc	STATUS,C
	goto	no_carry_sub24
		movf	dest+1,f
		btfsc	STATUS,Z
			decf	dest+2,f
		decf	dest+1,f
no_carry_sub24

	endm


; version avec 2 registres et 1 immdiat
; dest = reg - val
; reg est l'adresse du LSB de la source
; dest est l'adresse du LSB de la destination
sub24irr	macro reg, dest, val
	local	no_carry_sub24
	; MSB
	movlw	HIGHBYTE(val)
	subwf	reg+2,w
	movwf	dest+2

	; MID
	movlw	MIDBYTE(val)
	subwf	reg+1,w
	movwf	dest+1
	btfss	STATUS, C
		decf	dest+2,f

	; LSB
	movlw	LOWBYTE(val)
	subwf	reg,w
	movwf	dest
	btfsc	STATUS,C
	goto	no_carry_sub24
		movf	dest+1,f
		btfsc	STATUS,Z
			decf	dest+2,f
		decf	dest+1,f
no_carry_sub24
	endm

; version avec 1 registres et 1 immdiat
; reg = reg - val
; reg est l'adresse du LSB de la source
sub24ir	macro reg, val
	local	no_carry_sub24
	; MSB
	movlw	HIGHBYTE(val)
	subwf	reg+2,f

	; MID
	movlw	MIDBYTE(val)
	subwf	reg+1,f
	btfss	STATUS, C
		decf	reg+2,f

	; LSB
	movlw	LOWBYTE(val)
	subwf	reg,f
	btfsc	STATUS,C
	goto	no_carry_sub24
		movf	reg+1,f
		btfsc	STATUS,Z
			decf	reg+2,f
		decf	reg+1,f
no_carry_sub24
	endm



