; mtk/dispatch.inc
; version vendredi 9 novembre 2007
; version 14 novembre 2007 : changement de la structure des buffers de contexte
; version 29 novembre : sparation de mtk.inc
; version 2 dcembre : plus de fonctions, un peu plus compact, un peu plus clean
; version 20 fvrier : chang tmp_leaf* en MTK_leaf*
; version 2 mars : Macro_CALL spcifie aussi un flag pour conomiser qqs instructions.
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

	cblock ; utilis par le dispatcheur :
		wait_flags ; copie de INTCON, le temps de scanner la liste
		last_context_buffer ; bout de la liste chaine.
		current_context_buffer; adresse du contexte en cours d'excution
	endc

; flags d'optimisation : valeur=0 => dsactiv.
optimisation_goto13		set 1 ; critique quand la taille du code > 2K instr
optimisation_setstack	set 1 ; peut tre dangereux si mal utilis.
  ; Dans le doute, mettre  0, et tester un par un les macro_call xxx, 1

GOTO13bits	macro	adresse
	if (((adresse >> 13) != (($+1) >> 13)) || ((adresse >> 13) != (($+2) >> 13)) || (optimisation_goto13==0))
		if (adresse >> 8) != 0
			movlw	adresse >> 8
			movwf	PCLATH
		else
			clrf	PCLATH
		endif
	endif
	goto	adresse & 2047
	endm

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

; En raison de la structure interne des PIC,
; le MTK ne peut pas prserver la pile matrielle
; entre deux changements de threads.
; Il faut donc crer un systme d'appels "personalis",
; plus lourd et en soft, mais qui permet de changer
; de contexte sans perdre d'information.

; MCALL et MRETURN sont deux macros qui remplacent call
; et return. Elles peuvent travailler sur des adresses 13 bits.
; Cela simplifie donc les gros programmes car les threads
; peuvent tre logs dans les banques hautes de mmoire.

; La pile est pointe par FSR et accdde par INDF.
; Petit dtail pour aider  comprendre : la pile s'tend vers le haut.

; "amliorations" :
;        --> dsactives pour l'instant car bugs
; - retour : adresse de retour diffrente de l'instruction suivante.
;     utile pour conomiser un goto aprs le macro_call.
; - same_MSB : indique s'il faut mettre  jour l'adresse du MSB sur la pile
;     utiliser avec prcaution, mais gagne 2 instructions pour les appels conscutifs.

; Attention, Macro_CALL et Macro_RETURN dtruisent W !
Macro_CALL	macro	adresse, same_MSB
	local	label
	if (adresse < 1) ||(adresse > $)
		error "adresse de CALL invalide"
	endif
; sauve l'adresse de retour sur la "pile". Les LSB sont dpils en dernier, donc empils en premier
	movlw	label & 255 
	movwf	INDF
	incf	FSR, f
; L'adresse de retour est souvent proche de celle du prcdent appel, on peut sauter l'criture du MSB.
; A partir d'ici, il y a de 2  6 instructions jusqu'au label.
	if ((same_MSB==0) || (MTK_MSB_stack != (($+2)>>8)) || (MTK_MSB_stack != (($+6)>>8)) || (optimisation_setstack==0))
		movlw	label >> 8
		movwf	INDF
	endif
	incf	FSR, f
	GOTO13bits	adresse
label ; Macro_RETURN fera un goto ici.
MTK_MSB_stack	set label >> 8 ; pour aprs
	endm

; pareil mais spcifie en plus une adresse de retour
Macro_CALL_address	macro	adresse, retour, same_MSB
	if (adresse < 1) ||(adresse > $)
		error "adresse de CALL invalide"
	endif
; sauve l'adresse de retour sur la "pile". Les LSB sont dpils en dernier, donc empils en premier
	movlw	retour & 255 
	movwf	INDF
	incf	FSR, f
	if ((same_MSB==0) || (MTK_MSB_stack != ((retour)>>8)) || (optimisation_setstack==0))
		movlw	retour >> 8
		movwf	INDF
	endif
	incf	FSR, f
	GOTO13bits	adresse
MTK_MSB_stack	set retour >> 8 ; pour aprs
	endm


MTK_MSB_stack	set -1 ; initialise  rien


Macro_RETURN	macro
; dpile PCLATH :
	decf	FSR, f
	movfw	INDF
	movwf	PCLATH
; dpile PCL et branche :
	decf	FSR, f
	movfw	INDF
	movwf	PCL
	endm


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

; Chaque thread possde une zone de contexte.
; Tous ces buffers doivent tre en RAM bank0 ou bank1
; (l, ils sont par dfaut en banque 1)
; Attention ! on n'a que 96+80 octets pour tout le monde
; et chaque contexte bouffe au moins 8 octets
; sans compter les autres variables...
; Donc il est possible de mettre la pile en banques 2/3

; Structure du buffer de contexte :
; [ne pas changer l'ordre, a bousillerait tout le code !]
#define OffsetState	0 ; Status (runnable, waiting, ...)
#define OffsetVal	1 ; Timer value (ractive le thread si descend  0)
		; aussi utilis pour la liste chaine des smaphores,
		; les masques de rveil, et d'autres trucs encore...
#define OffsetNext	2 ; adresse du prochain contexte
#define OffsetSP	3 ; Stack Pointer (FSR)
#ifdef	STACK_HIGH
	#define OffsetSTATUS	4 ; 
	#define OffsetStack		5 ; premier octet de la pile
#else
	#define OffsetStack		4 ; premier octet de la pile
#endif

; valeurs dans les 3 LSB de [OffsetState]
#define	THREAD_ASLEEP		0	; ok
#define	THREAD_RUNNING		1	; ok
#define	THREAD_WAIT_MS		2	; ok
;#define	THREAD_WAIT_SIG		3	; pas encore fait
#define	THREAD_WAIT_BIT0	4	; ok
#define	THREAD_WAIT_BIT1	5	; ok
#define	THREAD_WAIT_BYTE0	6	; ok
#define	THREAD_WAIT_BYTENZ	7	; ok

; les bits 4  6 sont utiliss par THREAD_WAIT_BITx pour le n du bit  tester

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

; Alloue les contextes en partant des adresses les plus hautes :
NEXT_CONTEXT_ADDRESS	set LAST_CONTEXT_ADDRESS

ALLOCATE_CONTEXT		macro	context_name, stack_depth
NEXT_CONTEXT_ADDRESS	set		NEXT_CONTEXT_ADDRESS-(OffsetStack+(2*stack_depth))
	; on pourrait sauter  la banque prcdente mais c'est trop complexe pour l'instant.
	if NEXT_CONTEXT_ADDRESS < FIRST_CONTEXT_ADDRESS
		error "trop de contextes crs"
	endif
context_name			equ		NEXT_CONTEXT_ADDRESS
	endm

#ifdef	STACK_HIGH

; Alloue de l'espace pour la pile, hors de l'espace des buffers de contexte :
NEXT_STACK_ADDRESS	set LAST_STACK_ADDRESS

ALLOCATE_STACK		macro	stack_name, stack_depth
NEXT_STACK_ADDRESS	set		NEXT_STACK_ADDRESS-(2*stack_depth)
	if NEXT_STACK_ADDRESS < FIRST_STACK_ADDRESS
		error "trop de pilees crs"
	endif
stack_name			equ		NEXT_STACK_ADDRESS
	endm

#endif

; cette macro est insre dans la boucle d'initialisation de tous les buffers de contexte.
; elle calcule les adresses automatiquement en fonction du mode de fonctionnemnt,
; et vrifie que les paramtres sont corrects.

MACRO_init_thread macro	thread_base, StartIP, thread_status, thread_value, stack_address, stack_depth
#ifdef	STACK_HIGH
	ALLOCATE_CONTEXT	thread_base, 0
	ALLOCATE_STACK		stack_address, stack_depth
#else
	ALLOCATE_CONTEXT	thread_base, stack_depth
stack_address	set		thread_base+OffsetStack
#endif
	retlw	thread_base
	retlw	thread_status ; remplit le champ d'tat
	retlw	thread_value ; remplit VAL
#ifdef	STACK_HIGH
	retlw	(stack_address & 255)
	retlw	(stack_address >> 1) & 128
#endif
	retlw	StartIP&255; remplit la pile avec l'adresse  excuter
	retlw	StartIP>>8
#ifdef	PROTECT_STACK
	if stack_depth<2
		error "la pile doit au moins contenir 2 niveaux"
	endif
	retlw	stack_depth-1
#else
	if stack_depth<1
		error "la pile doit au moins contenir 1 niveau"
	endif
#endif
	endm

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; allocation de bits :



;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

; "appels systme"

CALL_MTK_YIELD	macro	retour
; nouvelle version encore plus compacte
	local	label
	if (retour!=0)
label	set	retour
	endif
; vrifier s'il faut vraiment changer PCLATH ? pour conomiser 1 ou 2 instructions 
	if (((($+5) >> 13) != (MTK_YIELD >> 13)) || (optimisation_goto13==0))
; on ne peut pas utiliser (label-1) car sa valeur n'est pas encore dfinie et dpend du calcul !
		if (MTK_YIELD >> 8) != 0
			movlw	MTK_YIELD >> 8
			movwf	PCLATH
		else
			clrf	PCLATH
		endif
	endif
	movlw	label>>8
	movwf	MTK_leaf1
	movlw	label&255
	goto	MTK_YIELD & 2047	
	if (retour==0)
; le return fera un goto ici.
label
	endif
	endm


; Utilis par un thread pour attendre sans bloquer tout le monde
; le dlai est donn en -peu-prs-millisecondes dans W
Macro_DELAY	macro	retour
	local	label
	if (retour!=0)
label	set	retour
	endif
	movwf	MTK_leaf2 ; sauve la valeur d'attente pour plus tard
	if (((($+5) >> 13) != (MTK_DELAY >> 13)) || (optimisation_goto13==0))
		if (MTK_DELAY >> 8) != 0
			movlw	MTK_DELAY >> 8
			movwf	PCLATH
		else
			clrf	PCLATH
		endif
	endif
	movlw	label>>8
	movwf	MTK_leaf1
	movlw	label&255
	goto	MTK_DELAY & 2047	
	if (retour==0)
; le return fera un goto ici.
label
	endif
	endm

; attend sur un octet ou un bit  1 ou 0
; l'adresse  tester est dans W
Macro_WAIT	macro type, bitoffset, retour
	local	label
	if (retour!=0)
label	set	retour
	endif
	movwf	MTK_leaf2 ; sauve l'adresse
	if (((($+7) >> 13) != (MTK_CHANGE_STATE >> 13)) || (optimisation_goto13==0))
		if (MTK_CHANGE_STATE >> 8) != 0
			movlw	MTK_CHANGE_STATE >> 8
			movwf	PCLATH
		else
			clrf	PCLATH
		endif
	endif
	movlw	type | ((bitoffset&7)<<4)
	movwf	MTK_leaf4
	movlw	label>>8
	movwf	MTK_leaf1
	movlw	label&255
	goto	MTK_CHANGE_STATE & 2047	
	if (retour==0)
; le return fera un goto ici.
label
	endif
	endm

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

; Les accs indirects sont  viter dans les threads,  moins de sauver FSR
; avant de l'utiliser. Macro_SaveFSR/Macro_RestoreFSR aide pour cela :

; versions courtes.
; - W est dtruit
; - ne pas utiliser dans une fonction appele par Macro_CALL ET par plusieurs threads diffrents
; (la mauvaise adresse risquerait d'tre utilis)
Macro_SaveFSR_address macro adresse
	movfw	FSR
	movwf	adresse
	endm

Macro_RestoreFSR_address macro adresse
	movfw	adresse
	movwf	FSR
	endm

; versions plus pratiques qui sauvent directement dans le buffer de contexte
; - W est toujours dtruit
; - le symbole current_context_base doit tre dfini au dbut du fichier source du thread.
;    (en cas d'oubli, c'est un autre thread qui va morfler)
; - a ne permet pas des appels rcursifs
; - quand le contexte est en banque 1, a risque de foirer
; mais c'est quand mme mieux dans une fonction leaf.
Macro_SaveFSR macro
	movfw	FSR
	movwf	current_context_base+OffsetSP
	endm

Macro_RestoreFSR macro
	movfw	FSR
	movwf	current_context_base+OffsetSP
	endm

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

