; drv16F887\LCD.asm
;  excuter en banque data 0
; utilise delay.asm/inc
; attention aux interactions avec le relais
; Configuration : 1 ligne de 16 caractres en 5x8 bits
;  (les bits N et F sont dfinis dans le .def)
; les gros problmes sont les temporisations et l'criture
; dans la deuxime moiti de l'cran.

; version 20080220 : gagn 1 niveau de pile HW en temporisant directement dans LCD_Strobe_E
; version 20080227 : clrlcd trop court

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; routine "call/return"
; les appels  S__Delay_1us sont supprims pour
; rduire l'occupation de la pile HW
S__LCD_Strobe_E
	bsf		LCD_E
	Macro_delay_1us ; ajoute 6 cycles d'attente
	bcf		LCD_E
	goto	$+1 ; 2 cycles, le return en ajoute 2,
	; et les instructions suivantes du programme font le reste
	return

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

; tmp_leaf1=adresse logique ( l'cran)
LS1__set_DDRAM_address_linear ; version appelable "long/macro", index dans tmp_leaf1
	movlw	2
	movwf	LCD_delay
	movfw	tmp_leaf1
	; fold !

; W = adresse en DDRAM
S1__set_DDRAM_address_linear
; corrige l'index si >=64
	movwf	cursor_index ; sauve pour de futurs accs avec display_string
	btfsc	cursor_index, 3 ; si l'adresse >8
	xorlw	64^8	; alors dplacer le MSB
	; fold !

; W=adresse dans la mmoire
S1__set_DDRAM_address_physical
	bcf		LCD_RS ; control transfer
	iorlw	LCD_Instr_DDRAMaddr
;	goto	S1__LCD_send_byte
; fold !

; ne pas oublier d'attendre ensuite, CALL_MTK_YIELD n'est pas suffisant !

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; envoie l'octet dans W en 2 parties, sans toucher aux autres bits
; (n'affecte ni RD3, ni RS, ni RW... donc utilisable
;  pour les instructions LCD et les donnes LCD)
S1__LCD_send_byte        ; caractre dans W : appel "short"/call
	movwf	tmp_leaf1
S1__LCD_send_byte_direct ; caractre dans tmp_leaf1 : appel "Long"/macro (car W est dtruit)
	bcf		LCD_RW	; criture
	movlw	0x0F
	andwf	Port_LCD, w
	movwf	tmp_leaf2

	; nibble haut
	movfw	tmp_leaf1
	andlw	0xF0
	iorwf	tmp_leaf2, w
	movwf	Port_LCD
	call	S__LCD_Strobe_E

	; nibble bas
	swapf	tmp_leaf1, w
	andlw	0xF0
	iorwf	tmp_leaf2, w
	movwf	Port_LCD
	call	S__LCD_Strobe_E

; en fonction du mode d'exploitation : delay_ms+return ou MTK_DELAY+MACRO_RETURN ?

	movf	LCD_delay,w
	btfsc	STATUS,Z
	return	; Si LCD_delay est  0 alors on retourne tout de suite
	; sinon on temporise et on retourne "long"
	Macro_DELAY	0
macro_return_lcd ; Macro_RETURN prend plus de place que juste return
	Macro_RETURN


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; W = caractre
; utilis lorsque le MTK est HS, ou assimil (init LCD)
; routine "call/return", ncessite 2 niveaux de pile
S2__LCD_send_data_init
	bsf		LCD_RS ; data transfer
; fold dans S2__LCD_send_byte_init
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; W = caractre
; utilis lorsque le MTK est HS, ou assimil
S2__LCD_send_byte_init
	call	S1__LCD_send_byte
	movlw	1
	call	S__Delay_ms
	return

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; routine d'init de l'afficheur, voir HD44780U.pdf fig. 26 p. 213
; attention : routine "call/return", qui dure trs longtemps
; cursor_type doit tre configur
S3__Reset_LCD
	clrf	LCD_delay ; s'assure que LCD_send_byte retourne directement sans tempo
	movlw	3
	movwf	tmp_leaf4
loop_reset_LCD
	; "Function set", interface 8 bits
		movlw	0x30
		movwf	Port_LCD
		call	S__LCD_Strobe_E
		movlw	20 ; "wait for more than 4,1ms"
		call	S__Delay_ms

	decfsz	tmp_leaf4,f
	goto	loop_reset_LCD

; "Function set", set interface to 4 bits
	movlw	0x20
	movwf	Port_LCD
	call	S__LCD_Strobe_E
	movlw	10
	call	S__Delay_ms

; l'interface est maintenant sur 4 bits,
; on peut utiliser les fonctions "normales"
; qui dcoupent l'octet en 2 nibbles.

; "Function set", interface 4 bits, taille fonte et nb de lignes
	movlw	LCD_Instr_FunctionSet|LCD_FORMAT|LCD_NB_LINES
	call	S2__LCD_send_byte_init

; "Display on" + contrle du curseur (voir LS2__change_cursor pour changer)
	movlw	LCD_Instr_DisplayControl | LCD_D_DisplayOn
	call	S2__LCD_send_byte_init

; "Cursor Shift"
    movlw	LCD_Instr_CursorShift | LCD_SC_CursorMove | LCD_RL_Right
	call	S2__LCD_send_byte_init

; "Display clear"
	movlw	LCD_Instr_ClearDisplay
	call	S2__LCD_send_byte_init

; "Entry mode set"
	movlw	LCD_Instr_EntryModeSet | LCD_ID_Increment
	call	S2__LCD_send_byte_init

; cursor address = 0 (pour de bon)
	clrf	cursor_index
; set cursor address to 0
	movlw	LCD_Instr_DDRAMaddr
	call	S2__LCD_send_byte_init

	return



;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; algo stupide de division par soustractions successives
; utilis par Display_int256
; base dans tmp_leaf2
; nombre  diviser dans convert_byte
; sortie : dans tmp_leaf1
; routine "macro", retourne/fold dans LCD_display_char
LS2__divide_int
	movlw	'0'	; prchargement
	movwf	tmp_leaf1
divide_loop	; nombre >= base ?
	movfw	tmp_leaf2
	subwf	convert_byte, w
	btfss	STATUS,C
	goto	strip_zero
	movwf	convert_byte
	incf	tmp_leaf1,f
	goto	divide_loop

strip_zero
	; rsultat = '0' ?
	movlw	'0'
	xorwf	tmp_leaf1,w
	btfsc	STATUS,Z
	goto	strip0
	; si ce n'est pas 0, alors mettre le flag
	bsf		leading0,0
	goto	LS2__LCD_display_char ; fold !!

strip0
	; faut-il remplacer le 0 par ' ' ?
	btfsc	leading0,0
	goto	LS2__LCD_display_char ; fold !!

	; remplacement :
	movlw	' '
	movwf	tmp_leaf1
;	goto	LS2__LCD_display_char
; fold !!

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; tmp_leaf1 = caractre  afficher
LS2__LCD_display_char
	bsf		LCD_RS ; data transfer
;	movfw	tmp_leaf1 ; simplifi avec S1__LCD_send_byte_direct
	movlw	1 	; attendre au moins 100us
	movwf	LCD_delay
	Macro_CALL	S1__LCD_send_byte_direct, 0

; incrmente le curseur
	incf	cursor_index,f
; dtecte le passage dans la deuxime moiti
	movfw	cursor_index
	xorlw	8
	btfss	STATUS,Z
	goto	macro_return_lcd  ; conomie de place oblige
; change l'adresse du curseur
	incf	LCD_delay,f ; delay=2
	movlw	64
	goto	S1__set_DDRAM_address_physical ; retour "long" par S1__LCD_send_byte
;	movlw	2
;	Macro_DELAY
;	goto	macro_return_lcd


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; arguments :
; string_low  : 8 LSB de l'adresse
; string_high : 5 MSB de l'adresse
; string en ASCIIZ (termine par un 0)
; W : impossible car Macro_CALL

; notes du 20080221 :
; - l'incrmentation est faite pour chaque appel, donc
;     on peut rappeler la fonction juste aprs son retour
;      pour afficher une chaine conscutive. conomie d'appel.
; - la fonction retourne si elle dpasse de l'cran.
;     donc on conomise aussi le 0 terminal.

L1S2__display_string
	call	call_char ; rcupre le prochain caractre
	movwf	tmp_leaf1
; restaurer PCLATH ! car l'adresse est certainement hors du bloc courant
	movlw	$>>8   ; et a risque de faire foirer le prochain call
	movwf	PCLATH

; incrmente le pointeur, version 16 bits
	incf	string_low,f
	btfsc	STATUS,Z
	incf	string_high,f

; si la valeur de retour est nulle, quitter.
	movf	tmp_leaf1,f
	btfsc	STATUS,Z
	goto	macro_return_lcd

; affiche le caractre
	Macro_CALL	LS2__LCD_display_char, 0

; quitter aussi si l'index dpasse l'cran
	btfsc	cursor_index, 4 ; si l'adresse >16
	goto	macro_return_lcd

	goto	L1S2__display_string ; reboucle

; pseudo-fonction qui permet de sauter  un truc qui fait un return
call_char
	movfw	string_high
call_char_2
	movwf	PCLATH
	movfw	string_low
	movwf	PCL ; saut de l'ange
; ATTENTION A RESTAURER PCLATH !!!

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; arguments :
; tmp_leaf1 : n du caractre
; tmp_leaf2 : nb d'octets  envoyer si non nul
; string_low  : 8 LSB de l'adresse de la fonte
; string_high : 5 MSB de l'adresse
L1S2__load_font
	; Si tmp_leaf2==0 alors convert_byte=7 sinon convert_byte=tmp_leaf2
	movfw	tmp_leaf2
	btfsc	STATUS,Z
	movlw	7
	movwf	convert_byte

	; fixe l'adresse en CGRAM
	bcf		LCD_RS ; control transfer
	movlw	2
	movwf	LCD_delay
	; ajuster l'index du caractre : bits 0-2 dans 5-3
	swapf	tmp_leaf1,f
	rrf		tmp_leaf1,f
	bsf		tmp_leaf1,6 ; bit "adresse CGRAM"
	Macro_CALL	S1__LCD_send_byte_direct, 0

	decf	LCD_delay,f ; delai=1ms
	bsf		LCD_RS ; data transfer

; envoie les octets
loop_load_font
		call	call_char ; rcupre le prochain caractre
		movwf	tmp_leaf1 ; affiche le caractre
		Macro_CALL	S1__LCD_send_byte_direct, 1
		; incrmente le pointeur, version 16 bits
		incf	string_low,f
		btfsc	STATUS,Z
		incf	string_high,f
	decfsz	convert_byte,f
	goto	loop_load_font

	call	call_char ; rcupre le dernier caractre
	goto	S1__LCD_send_byte ; retour par l


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; remet l'cran  0
LS2__clrlcd
	bcf		LCD_RS ; control transfer
	clrf	cursor_index
	movlw	5	; force le retour "long" par S1__LCD_send_byte
	movwf	LCD_delay
	movlw	LCD_Instr_ClearDisplay
	goto	S1__LCD_send_byte ; fold !

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; change le type de curseur (valeur dans tmp_leaf1)
; valeurs possibles : 
; LCD_C_CursorOff (0) / LCD_C_CursorOn (1<<1)
; LCD_B_CursorBlock	(0) / LCD_B_CursorBlink	(1<<0)
LS2__change_cursor
	bcf		LCD_RS ; control transfer
; "Display on" + contrle du curseur :
	movlw	LCD_Instr_DisplayControl | LCD_D_DisplayOn
	iorwf	tmp_leaf1,f ; ajoute les bits dsirs par le programme
	movlw	2	; force le retour "long" par S1__LCD_send_byte
	movwf	LCD_delay
	goto	S1__LCD_send_byte_direct ; fold !

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; convert_byte = valeur  afficher
; routine "macro", retourne au travers de LCD_display_char
L1S2_Display_int256
	bcf	leading0,0 ; peut tre omis si les 0 sont dsirs

; centaines
	movlw	100
	movwf	tmp_leaf2
	Macro_CALL	LS2__divide_int, 0

; dizaines
	movlw	10
	movwf	tmp_leaf2
	Macro_CALL	LS2__divide_int, 1

; units : plus rien  faire
	movfw	convert_byte ; entre 0 et 9
	addlw	'0' ; ajoute le code ASCII kivabien
	movwf	tmp_leaf1
	goto	LS2__LCD_display_char ; fold ! retourne au travers de LS2__LCD_display_char
;	goto	LS2__LCD_display_char_direct ; 1 instruction au lieu des 2 ci-dessus
;     ==> impossible car tmp_leaf1 doit tre crit pralablement
